/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/* SCCS ID: @(#)scan_defs.c	1.9 10/13/93 */
/*************************************************************************
 * FILE:      scan_defs.c
 * Author:    Claude Saunders
 * Purpose:   This module contains the data structures (and manipulation
 *            functions) which are built up when the server scans device
 *            definitions. The functions which actually scan device def-
 *            initions are here. The three primary data structures are 
 *            devicehead, classhead, and table. Functions for creating and 
 *            searching for device and class nodes are here. Standard hash 
 *            table access functions are here, although they are mostly just
 *            application specific wrappers which call functions in the
 *            hash table library gpdLib.c.
 *
 * Mods:      login mm/dd/yy: description
 *************************************************************************/

#include <stdlib.h>
#include <rpc/rpc.h>
#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include "dev.h"
#include "scan_defs.h"
#include "builderyacc.h"
#include "gpdLib2.h"

extern char *sys_errlist[];
extern int scanned_atomic;    /* defined in grammar definition dev.y */
extern char scan_error_msg[]; /* ditto */
extern int logerror(char*,char*);
extern int logserver(char*,char*);
extern int myParse();


/***************** Device Definition Data Structures ************************/

static PDEV devicehead = NULL;    /* singly linked list of device instances */
static PCLASS classhead = NULL;   /* singly linked list of device classes   */
static void *table = NULL;        /* hash table of device instances         */

/****************************************************************************/

/*************************************************************************
 * FUNCTION : deallocateClasslist()
 * PURPOSE  : Nulls and deallocates the device class list.
 * ARGS in  : none
 * ARGS out : none
 * GLOBAL   : module global classhead
 * RETURNS  : 0  if non-null list deallocated
 *            -1 if list already null
 ************************************************************************/
static int deallocateClasslist() {
  PCLASS pclass,npclass;
  PMSG pmsg,npmsg;

  if (classhead == NULL)
    return(-1);

  pclass = classhead;
  classhead = NULL;
  while (pclass != NULL) {      /* deallocate all CLASS nodes */
    pmsg = pclass->msgh;
    while (pmsg != NULL) {      /* deallocate all MSG nodes for that CLASS */
      npmsg = pmsg->next;
      free(pmsg);
      pmsg = npmsg;
    }
    npclass = pclass->next;
    free(pclass);
    pclass = npclass;
  }
  return(0);
}

/*************************************************************************
 * FUNCTION : printClasslist()
 * PURPOSE  : Prints out current contents of device class list.
 *            (not all data, but enough for debugging)
 * ARGS in  : none
 * ARGS out : none
 * GLOBAL   : module global classhead
 * RETURNS  : 0
 ************************************************************************/
static int printClasslist() {
  PCLASS pclass;
  PMSG pmsg;

  pclass = classhead;
  while (pclass != NULL) {
    pmsg = pclass->msgh;
    fprintf(stderr,"class %s\n",pclass->name);
    while (pmsg != NULL) {
      fprintf(stderr,"    msg %s\n",pmsg->name);
      pmsg = pmsg->next;
    }
    pclass = pclass->next;
  }
  return(0);
}

/*************************************************************************
 * FUNCTION : allocateClass()
 * PURPOSE  : Appends a CLASS node to the device class list. A ptr to this
 *            new node is returned.
 * ARGS in  : none
 * ARGS out : none
 * GLOBAL   : module global classhead
 * RETURNS  : ptr to new CLASS node
 ************************************************************************/
PCLASS allocateClass() {
  PCLASS pclass, newclass;

  if (classhead == NULL) {
    classhead = (PCLASS)malloc(sizeof(CLASS));
    classhead->next = NULL;
    classhead->msgh = NULL;     /* no MSG nodes yet */
    return(classhead);
  }
  else {
    pclass = classhead;
    while (pclass->next != NULL)
      pclass = pclass->next;
    newclass = (PCLASS)malloc(sizeof(CLASS));
    newclass->next = NULL;
    newclass->msgh = NULL;
    pclass->next = newclass;
    return(newclass);
  }
}

/*************************************************************************
 * FUNCTION : findClass()
 * PURPOSE  : Search device class list for classname. Return ptr to the
 *            matching CLASS node. 
 * ARGS in  : classname: string containing class name
 * ARGS out : none
 * GLOBAL   : module global classhead
 * RETURNS  : NULL if classname doesn't match with any node in class list,
 *            ptr to CLASS node otherwise
 ************************************************************************/
PCLASS findClass(classname)
     char *classname;
{
  PCLASS pclass;

  if (classhead == NULL)
    return(NULL);
  else {
    pclass = classhead;
    while (pclass != NULL)
      if (strcmp(classname,pclass->name))
	pclass = pclass->next;
      else
	break;
    return(pclass);
  }
}

/*************************************************************************
 * FUNCTION : allocateDev()
 * PURPOSE  : Appends a DEV node to the node pointed at by pdev. This
 *            new DEV node is returned.
 * ARGS in  : pdev: ptr to DEV node
 * ARGS out : none
 * GLOBAL   : appends a node to module global devicehead
 * RETURNS  : ptr to new DEV node
 ************************************************************************/
PDEV allocateDev(pdev) 
     PDEV pdev;
{
  PDEV newdev;

  newdev = (PDEV)malloc(sizeof(DEV));
  newdev->next = NULL;
  newdev->pvmsgh = NULL;   /* ptr to list of supported messages/pvnames */

  if (pdev != NULL)
    pdev->next = newdev;

  return(newdev);
}

/************************************************************************/
/************** Device Hash Table Access Functions **********************/
/************************************************************************/

/*************************************************************************
 * FUNCTION : dev_find()
 * PURPOSE  : Locates entry in hash table for devname. The user field of
 *            the hash table entry is returned. In this case, a ptr to
 *            the DEV node corresponding to devname.
 *
 * ARGS in  : devname: string representing device name
 * ARGS out : none
 * GLOBAL   : uses gpdFind function to perform find on table
 * RETURNS  : NULL if no entry in table for devname,
 *            ptr to DEV node otherwise
 ************************************************************************/
PDEV dev_find(devname)
     char *devname;
{
  GPDENTRY *entry;
  entry = gpdFind(table,devname," ");
  if (entry == NULL)
    return((PDEV)NULL);
  else
    return((PDEV)entry->userPvt);
}

/*************************************************************************
 * FUNCTION : dev_add()
 * PURPOSE  : Adds a new entry to hash table for devname, but only if it
 *            isn't there already. The user field is initialized with a
 *            ptr to a newly allocated DEV node. 
 *
 * ARGS in  : devname: name of new device to add to hash table
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : ptr to DEV node
 ************************************************************************/
PDEV dev_add(devname)
     char *devname;
{
  PDEV pdev;
  GPDENTRY *entry;

  entry = gpdFind(table,devname," ");
  if (entry != NULL)
    return((PDEV)entry->userPvt);
  pdev = allocateDev((PDEV)NULL);
  entry = gpdAdd(table,devname," ");
  entry->userPvt = (void *)pdev;
  return(pdev);
}

/*************************************************************************
 * FUNCTION : dev_delete()
 * PURPOSE  : Removes entry from hash table for devname. The gpdDelete
 *            function is used (which in turn takes a ptr to a function
 *            for deleting the user field of the hash table entry).
 *
 * ARGS in  : devname: name of table entry for deletion
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : nothing
 ************************************************************************/
void dev_delete(devname)
     char *devname;
{
  PDEV pdev;
  GPDENTRY *entry;

  entry = gpdFind(table,devname," ");
  if (entry != NULL) {
    pdev = (PDEV)entry->userPvt;
    if (pdev->type == ATOMIC)
      gpdDelete(table,devname,"",freeDev); 
    else
      gpdDelete(table,devname,"",freeCDev);
  }
  return;
}

/*************************************************************************
 * FUNCTION : dev_next()
 * PURPOSE  : Used to do a linear scan of the hash table, returning one
 *            entry each time it is called. The op argument is used to
 *            request the next entry, or to rewind the scan.
 *
 * ARGS in  : op: 0 - reset the scan to the beginning of the table
 *                1 - get next entry in hash table
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : NULL if no more entries in table (or op was 0),
 *            ptr to DEV node from user field of table entry otherwise
 ************************************************************************/
PDEV dev_next(int op)
{
  GPDENTRY *entry;

  if (op == 0) {
    gpdNext(table,0);    /* use gpdNext function to perform reset */
    return(NULL);
  }
  else {
    entry = gpdNext(table,1);   /* get next entry */
    if (entry == NULL)
      return(NULL);
    else
      return ((PDEV)entry->userPvt);   /* return user field */
  }
}

/******************* end Hash Table Functions ****************************/

/*************************************************************************
 * FUNCTION : allocatePVMsg()
 * PURPOSE  : Appends a new PVMSG node to the singly linked pvmsgh list
 *            hanging off each DEV node. 
 *
 * ARGS in  : pdev: ptr to DEV node to which a new PVMSG node is added
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : ptr to newly allocated PVMSG node
 ************************************************************************/
PPVMSG allocatePVMsg(pdev)
  PDEV pdev;
{
  PPVMSG ppvmsg,newpvmsg;

  if (pdev->pvmsgh == NULL) {
    pdev->pvmsgh = (PPVMSG)malloc(sizeof(PVMSG));
    pdev->pvmsgh->next = NULL;
    return(pdev->pvmsgh);
  }
  else {
    ppvmsg = pdev->pvmsgh;
    while (ppvmsg->next != NULL) 
      ppvmsg = ppvmsg->next;
    newpvmsg = (PPVMSG)malloc(sizeof(PVMSG));
    newpvmsg->next = NULL;
    ppvmsg->next = newpvmsg;
    return(newpvmsg);
  }
}

/*************************************************************************
 * FUNCTION : allocateMsg()
 * PURPOSE  : Appends a new MSG node to the singly linked msgh list hanging
 *            off of each CLASS node.
 *
 * ARGS in  : pclass: ptr to CLASS node to which a new MSG node is added
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : ptr to newly allocated MSG node
 ************************************************************************/
PMSG allocateMsg(pclass)
     PCLASS pclass;
{
  PMSG pmsg,newmsg;

  if (pclass->msgh == NULL) {
    pclass->msgh = (PMSG)malloc(sizeof(MSG));
    pclass->msgh->next = NULL;
    return(pclass->msgh);
  }
  else {
    pmsg = pclass->msgh;
    while (pmsg->next != NULL) 
      pmsg = pmsg->next;
    newmsg = (PMSG)malloc(sizeof(MSG));
    newmsg->next = NULL;
    pmsg->next = newmsg;
    return(newmsg);
  }
}

/*************************************************************************
 * FUNCTION : findMsg()
 * PURPOSE  : Searches linked list of MSG nodes attached to a given CLASS
 *            node for one which matches msgname. 
 *
 * ARGS in  : pclass: ptr to class node which contains list of MSG nodes
 *            msgname: search for this string in the name field of MSG nodes
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : ptr to MSG node if match found,
 *            NULL otherwise
 ************************************************************************/
PMSG findMsg(pclass,msgname)
     PCLASS pclass;
     char *msgname;
{
  PMSG pmsg;

  if (pclass == NULL)
    return(NULL);
  else {
    pmsg = pclass->msgh;
    while (pmsg != NULL)
      if (strcmp(msgname,pmsg->name))
	pmsg = pmsg->next;
      else
	break;
    return(pmsg);
  }
}

/*************************************************************************
 * FUNCTION : freeDev()
 * PURPOSE  : Function passed to gpdDelete. It serves to free up the
 *            DEV data structure for an atomic device in the user field
 *            of the hash table entry.
 *
 * ARGS in  : userPvt: user field of hash table entry
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : nothing
 ************************************************************************/
void freeDev(void *userPvt)
{
  PDEV pdev, npdev;
  PPVMSG ppvmsg, nppvmsg;

  pdev = (PDEV)userPvt;
  while (pdev != NULL) {        /* for each device */
    ppvmsg = pdev->pvmsgh;
    while (ppvmsg != NULL) {    /* for each message of device */
      nppvmsg = ppvmsg->next;
      free(ppvmsg);
      ppvmsg = nppvmsg;
    }
    npdev = pdev->next;
    free(pdev);
    pdev = npdev;
  }
}

/*************************************************************************
 * FUNCTION : freeCDev()
 * PURPOSE  : Function passed to gpdDelete. It serves to free up the 
 *            DEV data structure for a composite device in the user
 *            field of the hash table entry.
 *
 * ARGS in  : userPvt: user field of hash table entry
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : nothing
 ************************************************************************/
void freeCDev(void *userPvt)
{
  PDEV pdev, npdev;
 
  pdev = (PDEV)userPvt;
  while (pdev != NULL) {     /* for each device */
    npdev = pdev->next;
    free(pdev);
    pdev = npdev;
  }
}

/*************************************************************************
 * FUNCTION : scan_defs()
 * PURPOSE  : Performs server startup scan of device definition files.
 *            The DEV_DEF_PATH environment variable should specify what
 *            directory contains the standard set of device definitions
 *            the server should know about. Class definitions are parsed
 *            first (.cl files), followed by atomic definitions (.at files),
 *            and lastly any composite definitions (.co files). The parsing
 *            of files is accomplished by pointing stdin to the open file,
 *            and then calling the yyparse() function (generated by yacc).
 *            The various data structures are built up while yyparse() is
 *            running (see dev.y).
 *
 * ARGS in  : none
 * ARGS out : none
 * GLOBAL   : none
 * RETURNS  : 0  - scanning completed successfully
 *            -1 - error during scanning (syntactical or semantic)
 ************************************************************************/
int scan_defs()
{
  int status = -1;    /* return from yyparse() 0 all ok, syntax err otherwise*/
  char buf[255];
  char dirpath[255];
  FILE *fp;
  char *path;
  DIR *dirp;
  struct dirent *d;
  int classfound = 0;   /* 1 means at least one class definition was found */
  int atdevfound = 0;   /* 1 means at least one atomic definition was found */

  path = getenv("DEV_DEF_PATH");
  if (path == NULL)
    path = strdup(".");  /* default to current directory */

  dirp = opendir(path);
  if (dirp == NULL) {
    sprintf(buf,"can't open path: %s",path);
    logerror("init",buf);
    perror("");
    return(-1);
  }
  
  gpdFreeMem(table,freeDev); /* free up devices in hash table */
  gpdInitPvt(&table);        /* init new hash table */
  deallocateClasslist();     /* free up class definitions */

  while (d = readdir(dirp)) {
    if (!strcmp(".cl",(d->d_name+strlen(d->d_name)-3))) {
      dirpath[0] = '\0';
      strcpy(dirpath,path);
      strcat(dirpath,"/");
      strcat(dirpath,d->d_name);
      sprintf(buf,"file %s",dirpath);
      logserver("class parse",buf);
      if ((fp = freopen(dirpath,"r",stdin)) == NULL) {  /* set stdin to file */
	sprintf(buf,"can't open %s",dirpath);
	logerror("init",buf);
	perror("");
      }
      else {
	status = myParse();
	if (status == 0) classfound = 1;
	fclose(fp);
      }
      logserver("class parse","complete");
    }
  }
  if (!classfound) {
    logerror("init","no valid class defs found");
    closedir(dirp);
    return(-1);
  }
  rewinddir(dirp);
  while (d = readdir(dirp)) {
    if (!strcmp(".at",(d->d_name+strlen(d->d_name)-3))) {
      dirpath[0] = '\0';
      strcpy(dirpath,path);
      strcat(dirpath,"/");
      strcat(dirpath,d->d_name);
      sprintf(buf,"file %s",dirpath);
      logserver("atomic parse",buf);
      if ((fp = freopen(dirpath,"r",stdin)) == NULL) {
	sprintf(buf,"can't open %s",dirpath);
	logerror("init",buf);
	perror("");
      }
      else {
	status = myParse();
	if (status == 0) atdevfound = 1;
	fclose(fp);
      }
      logserver("atomic parse","complete");
    }
  }
  if (!atdevfound) {
    logerror("init","no valid atomic defs found");
    closedir(dirp);
    return(-1);
  }
  scanned_atomic = 1;
  rewinddir(dirp);
  while (d = readdir(dirp)) {
    if (!strcmp(".co",(d->d_name+strlen(d->d_name)-3))) {
      dirpath[0] = '\0';
      strcpy(dirpath,path);
      strcat(dirpath,"/");
      strcat(dirpath,d->d_name);
      sprintf(buf,"file %s",dirpath);
      logserver("composite parse",buf);
      if ((fp = freopen(dirpath,"r",stdin)) == NULL) {
	sprintf(buf,"can't open %s",dirpath);
	logerror("init",buf);
	perror("");
      }
      else {
	status = myParse();
	fclose(fp);
      }
      logserver("composite parse","complete");
    }
  }
  closedir(dirp);
  return(0);
}

/*************************************************************************
 * FUNCTION : scan_composite_defs_1()
 * PURPOSE  : This function implements the RPC SCAN_COMPOSITE_DEFS service.
 *            Given a path to a composite device definition file (.cl file),
 *            parse it, defining and/or redefining composite devices. Any
 *            errors encountered are returned as a string.
 *
 * ARGS in  : path: path to ascii composite device definition file
 * ARGS out : none
 * GLOBAL   : accesses global scan_error_msg which is in dev.y
 * RETURNS  : zero length string if no error, err message otherwise
 ************************************************************************/
char **scan_composite_defs_1(char **path)
{
  int status;
  static char *scan_msg;
  FILE *fp;
  char dirpath[255];
  char buf[255];

  status = 0;
  strcpy(dirpath,*path);

  xdr_free(xdr_string,(char *)&scan_msg);
  scan_msg = malloc(255);

  if ((fp = freopen(dirpath,"r",stdin)) == NULL) {
    sprintf(buf,"can't open %s",dirpath);
    logerror("composite parse",buf);
    perror("");
    strcpy(scan_msg,buf);
    strcat(scan_msg,": ");
    strcat(scan_msg,sys_errlist[errno]);
    return(&scan_msg);
  }
  status = myParse();
  fclose(fp);
  if (status != 0)
    strcpy(scan_msg,scan_error_msg);  /* copy err msg from parsing (dev.y) */
  else
    strcpy(scan_msg,"");

  return(&scan_msg);
}

